@@ -11,9 +11,8 @@
 namespace foxglove_bridge {
 
 foxglove::Parameter fromRosParam(const std::string& name, const XmlRpc::XmlRpcValue& value);
+XmlRpc::XmlRpcValue toRosParam(const foxglove::ParameterValue& param);
 
 std::vector<std::regex> parseRegexPatterns(const std::vector<std::string>& strings);
 
-bool isWhitelisted(const std::string& name, const std::vector<std::regex>& regexPatterns);
-
 }  // namespace foxglove_bridge
